"""
Text anonymization module
"""

from __future__ import annotations
from .analyzer import PersonalInfo
from datetime import datetime
from . import analyzer
from . import mask
from . import anotate
import json

class Anonymizer:
    """Anonymization class based on strategies formating"""

    def __init__(self):
        # available strategies
        self.ANALYZERS = {
            "regex": analyzer.RegexStrategy(),
            "pii": analyzer.PiiStrategy(),
            "lossy": analyzer.LossyStrategy(),
        }

        # available masks
        self.MASKS = {
            "placeholder": mask.PlaceholderStrategy(),
            "fake": mask.FakeStrategy(),
            "hash": mask.HashStrategy(),
            "hide": mask.HideStrategy(),
        }

        # available annotator
        self.ANNOTATORS = {
            "standoff": anotate.StandoffStrategy(),
            "doccano": anotate.DoccanoStrategy(),
            "uimacas": anotate.UimaCasStrategy(),
        }

        self._infos = None
        self._position = []
        self._mask = mask.PlaceholderStrategy()
        self._analyzers = []
        self._annotator = None

    def open_text_file(self, path: str) -> str:
        """
        Open input txt file

        :param path: path of the input txt file
        :returns: file content
        :raises FileExistsError: if given file not found
        """
        try:
            with open(path, "r") as f:
                content = f.read()
            return content
        except FileNotFoundError as e:
            print(e)
            raise

    def open_json_file(self, path: str) -> str:
        """
        Open input json file for personal infos

        :param path: path of the json file
        :returns: file content
        :raises FileExistsError: if given file not found
        """
        try:
            with open(path, "r") as f:
                data = json.load(f)
            return data
        except FileNotFoundError as e:
            print(e)
            raise

    def set_info(self, infos: PersonalInfo):
        """
        Set personal info

        :param infos: PersonalInfo
        """
        self._infos = infos
        return infos

    def set_info_from_dict(self, **kwargs):
        """
        Set dict to PersonalInfo Class

        :param infos: dict with all the Personal info values

        """
        clean_data = {
            k: (
                ""
                if v is None
                else v.strftime("%Y-%m-%d")
                if isinstance(v, datetime)
                else v
            )
            for k, v in kwargs.items()
        }
        info_obj = PersonalInfo(**clean_data)
        self._infos = info_obj
        return info_obj

    def add_analyzer(self, name: str):
        """
        Add analyser

        :param name: analyzer used for anonymisation

        """
        if name in self.ANALYZERS:
            analyzer = self.ANALYZERS.get(name)
            if analyzer not in self._analyzers:
                self._analyzers.append(analyzer)
        else:
            raise Exception(f"{name} analyzer doesn't exist")

    def set_mask(self, name: str):
        """
        Set masks

        :param name: wanted mask
        """
        if name in self.MASKS:
            self._mask = self.MASKS.get(name)

        else:
            raise Exception(f"{name} mask doesn't exist")

    def set_annotator(self, name: str):
        """
        Set annotator

        :param name: wanted annotator
        """
        if name in self.ANNOTATORS:
            self._annotator = self.ANNOTATORS.get(name)
        else:
            raise Exception(f"{name} annotator doesn't exist")

    def anonymize(self, text: str, infos: PersonalInfo = None) -> str:
        """
        Global function to anonymise a text base on the choosen strategies

        :param text: text to anonymize
        :returns: anonimized text
        """
        if not text:
            return "NaN"
        resolved_infos = infos if infos is not None else self._infos
        anonymized_text = text
        for strategy in self._analyzers:
            spans = strategy.analyze(text=anonymized_text, info=resolved_infos)
            anonymized_text = self._mask.mask(anonymized_text, spans)
        return anonymized_text

    def annotate(self, text: str) -> str:
        """
        Global function to annotate a text base on the choosen strategies

        :param text: text to annotate
        :returns: annonated text
        """
        spans = {}
        for strategy in self._analyzers:
            strategy.info = self._infos
            span = strategy.analyze(text=text)
            spans.update(span)
        if self._annotator:
            annotated_text = self._annotator.annotate(text, spans)
        return annotated_text



from incognito_anonymizer import Anonymizer
from incognito_anonymizer import analyzer
from incognito_anonymizer import mask
from incognito_anonymizer import PersonalInfo
from datetime import datetime
import pytest

dataset_regex = {
    "phone": ("tél: 0651565600", "tél: <NUMBER>"),
    "phone2": ("tél: 06 51 56 56 00", "tél: <NUMBER>"),
    "email": ("email : joe.lafripouille@chu-brest.fr", "email : <EMAIL>"),
    "email_2": ("email : Joe.LaFripouille@chu-brest.fr", "email : <EMAIL>"),
    "email_3": ("bob@gmail.com", "<EMAIL>"),
    "email_4": ("Joe.LaFripouille@chu-****.fr", "<EMAIL>"),
    "nir": ("nir : 164064308898823", "nir : <NUMBER>"),
    "nir_space": ("nir : 1 64 06 43 088 988 (23)", "nir : <NUMBER> (23)"),
    "NOM_Prenom": ("name : DUPONT Jean", "name : DUPONT Jean"),
    "Prenom_NOM": ("name : Jean DUPONT", "name : Jean DUPONT"),
    "Nom_compose_Prenom": ("name : De La Fontaine Jean", "name : De La Fontaine Jean"),
    "NOM-NOM_Prenom": ("name : DE-TROIS Jean", "name : DE-TROIS Jean"),
    "NOM_accent_Prénom": ("Monsieur JOÉÇ KAKŸÇ", "Monsieur <NAME>"),
    "NOM_accent_prénom": ("Monsieur JOÉÇ Poçèé", "Monsieur <NAME>"),
    "P._NOM": ("J. Jean", "J. Jean"),
    "Monsieur_NOM_Prenom": ("Monsieur KEAN Jean", "Monsieur <NAME>"),
    "Monsieur_NOM_Prenom_DOUBLE": (
        "Monsieur KEAN KEZAN Jean-Baptiste",
        "Monsieur <NAME>",
    ),
    "INTERNE_NOM-NOM_Prenom": (
        "name : Interne : DE-TROIS Jean",
        "name : Interne : <NAME>",
    ),
    "Titre_Interne": ("Interne", "Interne"),
    "Prenom_:_Prenom": ("Prenom : Georgette", "Prenom : <NAME>"),
    "Docteur_NOM_Prenom": ("Docteur DUPONT Jean", "Docteur <NAME>"),
    "Docteur_PRENOM_NOM": ("Docteur DUPONT JEAN", "Docteur <NAME>"),
    "Docteur\n_PRENOM_NOM": ("Docteur\nDUPONT JEAN", "Docteur\n<NAME>"),
    "Monsieur_P._NOM": ("Monsieur J. Jean", "Monsieur <NAME>"),
    "Monsieur_P._NOM_MAJUSCULE": ("Monsieur J. JEAN", "Monsieur <NAME>"),
    "Monsieur_P._NOM_apostrophe": ("Monsieur J. L'Jean", "Monsieur <NAME>"),
    "Monsieur_P._NOM_apostrophe_MAJ": ("Monsieur J. L'JEAN", "Monsieur <NAME>"),
    "Dr_NOM_Prenom": ("Dr LECLERC Charle", "Dr <NAME>"),
    "Dr_Prenom_NOM": ("Dr Charle LECLERC", "Dr <NAME>"),
    "Dr_P._P._NOM": ("Dr J.F. LECLERC", "Dr <NAME>"),
    "Dr_P._P._Nom": ("Dr J.F. Laclerc", "Dr <NAME>"),
    "Dr_P._P._NOM_2": ("Dr J. LECLERC", "Dr <NAME>"),
    "Dr_P._P._NOM_3": ("Dr J. LECLERC", "Dr <NAME>"),
    "Professeur_P._de_NOM": ("Professeur L. de LALALAND", "Professeur <NAME>"),
    "Professeur_P._du_NOM": ("Professeur L. du LALALAND", "Professeur <NAME>"),
    "Professeur_P._DE_NOM": ("Professeur L. DE LALALAND", "Professeur <NAME>"),
    "Professeur_P._DU_NOM": ("Professeur L. DU LALALAND", "Professeur <NAME>"),
    "Chef_de_service_P._NOM": ("Chef de service L. LARIDE", "Chef de service <NAME>"),
    "Cheffe_de_service_P._NOM": (
        "Cheffe de service L. LARIDE",
        "Cheffe de service <NAME>",
    ),
    "Mme_P._NOM": ("Mme C. CCCCC", "Mme <NAME>"),
    "DR._NOM": ("DR. LECLERC", "DR. <NAME>"),
    "PR_NAME": ("PR ABGRAL RONAN", "PR <NAME>"),
    "Interne_NOM_Prenom": ("Interne JEAN Jean", "Interne <NAME>"),
    "Externe_NOM_Prenom": ("Externe JEAN Jean", "Externe <NAME>"),
    "nom_phone": (
        "Monsieur JEAN Lasalle, tél : 0647482884",
        "Monsieur <NAME>, tél : <NUMBER>",
    ),
    "double_nom": (
        "Monsieur JEAN Jean, Docteur Jeanj JEAN, Madame JEANNE Jean",
        "Monsieur <NAME>, Docteur <NAME>, Madame <NAME>",
    ),
    "test": (
        "Bonjour Monsieur JEAN Jean, voici son numéro : 0606060606 et son email jean.jean@gmail.fr",
        "Bonjour Monsieur <NAME>, voici son numéro : <NUMBER> et son email <EMAIL>",
    ),
    "née_madame": ("Madame DUPONT Mariane née MORGAT", "Madame <NAME> née <NAME>"),
    "née_madame_2": ("Nom : DUPONT Mariane née MORGAT", "Nom : <NAME> née <NAME>"),
    "né_monsieur": ("Monsieur J. Jean né LA RUE", "Monsieur <NAME> né <NAME>"),
    "test_None": (None, "NaN"),
    "Prof_NOM_PRENOM": ("Professeur JEAN JEAN", "Professeur <NAME>"),
    "Profe_NOM_PRENOM": ("Professeure JEAN JEAN", "Professeure <NAME>"),
    "INT_NOM_PRENOM": ("INT JEAN JEAN", "INT <NAME>"),
    "Date_/": ("01/12/2000", "<DATE>"),
    "Date_-": ("01-12-2000", "<DATE>"),
    "Date_-": ("01-12-2000", "<DATE>"),
    "Date_incorrect": ("12-22-2000", "12-22-2000"),
    "Date_phrase": ("Brest, le 01/01/2000", "Brest, le <DATE>"),
    "adresse_2": ("155 rue de Brest, 29820, Guilers", "<ADRESSE>"),
    "adresse_3": ("28 RUE DU CHATEAU", "<ADRESSE>"),
    "adresse_4": ("20, bis rue de la Plage", "20, bis <ADRESSE>"),
    "date_mois": ("8 juillet 2020", "<DATE>"),
    "date_fourchette": ("du 15 au 24 octobre 2015", "du 15 au <DATE>"),
    "ville_date": ("BREST, le 4 Juin 2015", "BREST, le <DATE>"),
    "date_.": ("née le 27.05.31", "née le <DATE>"),
    "zip_name": ("29760 PENMARCH", "<ADRESSE>"),
    "zip_name_2": ("29270 CARHAIX PLOUGUER", "<ADRESSE>"),
    "zip_name_3": ("29270 AIX-EN-PROVENCE", "<ADRESSE>"),
}


datas_regex = list(dataset_regex.values())
ids_regex = list(dataset_regex.keys())


@pytest.mark.parametrize("input,output", datas_regex, ids=ids_regex)
def test_regex_strategie(input, output):

    ano = Anonymizer()
    ano.add_analyzer("regex")
    ano.set_mask("placeholder")
    assert ano.anonymize(input) == output


infos = {
    "first_name": "Lea",
    "last_name": "Jungels",
    "birth_name": "",
    "birthdate": datetime(1992, 9, 22, 0, 0, 0),
    "ipp": "0987654321",
    "postal_code": "01000",
    "adress": "",
}
dataset_pii = {
    "Nom_Prenom_PII": ("Léa Jungels", "<NAME> <NAME>"),
    "Date_IPP_Postal": ("22/09/1992 0987654321 01000", "<DATE> <IPP> <CODE_POSTAL>"),
    "DN": ("DN : 22/09/1992 ", "DN : <DATE> "),
}

datas_pii = list(dataset_pii.values())
ids_pii = list(dataset_pii.keys())


@pytest.mark.parametrize("input,output", datas_pii, ids=ids_pii)
def test_pii_strategie(input, output):
    ano = Anonymizer()
    ano.set_info_from_dict(**infos)
    ano.add_analyzer("pii")
    ano.set_mask("placeholder")
    assert ano.anonymize(input) == output


def test_anaylser_not_implemented_error():
    with pytest.raises(NotImplementedError):
        analyzer.AnalyzerStrategy.analyze(text="test")


def test_mask_not_implemented_error():
    with pytest.raises(NotImplementedError):
        mask.Strategy.mask("test", coordinate=((10, 10), "<TEST>"))


def test_add_analyser_error():
    ano = Anonymizer()
    with pytest.raises(Exception, match="test analyzer doesn't exist"):
        ano.add_analyzer("test")


def test_set_mask_error():
    ano = Anonymizer()
    with pytest.raises(Exception):
        ano.set_mask("test")


def test_set_annotator_error():
    ano = Anonymizer()
    with pytest.raises(Exception):
        ano.set_annotator("test")


dataset_regex = {
    "NOM_Prenom": ("name : DUPONT Jean", "name : <NAME>"),
    "Prenom_NOM": ("name : Jean DUPONT", "name : <NAME>"),
    "Prenom_NOM_2": ("Karine LACUT", "<NAME>"),
    "Prenom_NOM_3": ("   Jean DUPONT", "   <NAME>"),
    "N._Prenom": ("name : D. Jean", "name : <NAME>"),
    "N.-N._Prenom": ("name: D.D. Jean", "name: <NAME>"),
    "N_Prenom": ("L Johan", "<NAME>"),
    "NOM_Prenom_Composee": ("LARIDE Jean Philippe", "<NAME>"),
    "NOM_Prenom_Composee_2": ("LA RIDE Jean François", "<NAME>"),
    "Prenom_NOM_appostrophe": ("Jean HOC'H", "<NAME>"),
}


datas_regex = list(dataset_regex.values())
ids_regex = list(dataset_regex.keys())


@pytest.mark.parametrize("input,output", datas_regex, ids=ids_regex)
def test_lossy_strategie(input, output):

    ano = Anonymizer()
    ano.add_analyzer("lossy")
    ano.set_mask("placeholder")
    assert ano.anonymize(input) == output


dataset_regex = {
    "NOM_Prenom": ("Jeanne ALLEE", "<NAME>"),
    "word_test": ("l'incapacité de marcher par elle même", "l'incapacité de marcher par elle même"),
    "Monsieur_NOM_Prenom": ("Monsieur VIDAL Jean", "Monsieur <NAME>"),
    "medical_sentense": ("Voie orale, pendant 1 Mois", "Voie orale, pendant 1 Mois"),
}


datas_regex = list(dataset_regex.values())
ids_regex = list(dataset_regex.keys())


@pytest.mark.parametrize("input,output", datas_regex, ids=ids_regex)
def test_full_strategies(input, output):

    ano = Anonymizer()
    ano.add_analyzer("regex")
    ano.add_analyzer("lossy")
    ano.set_mask("placeholder")
    assert ano.anonymize(input) == output



from typing import Dict, List, Tuple
from cassis import Cas
import json

"""
    Classes to annotate the word at the given coordinates
"""


class Strategy:
    def annotate(text, coordinate: Dict[List[Tuple], str]):  # pragma: no cover
        raise NotImplementedError()


class StandoffStrategy(Strategy):
    """Generate BRAT-style standoff annotations"""

    def annotate(self, text, coordinate: Dict[List[Tuple], str]):
        lines = []
        tid = 1
        for coord_group, label in coordinate.items():
            label = label.strip("<>")
            for (start, end) in coord_group:
                span_text = text[start:end].replace("\n", " ")
                lines.append(f"T{tid}\t{label} {start} {end}\t{span_text}")
                tid += 1
        return "\n".join(lines)


class DoccanoStrategy(Strategy):
    def annotate(self, text, coordinate: Dict[List[Tuple], str]):
        """Generate Doccano jsonl format annotations"""
        label_data = []
        for spans, label in coordinate.items():
            label = label.strip("<>")
            for (start, end) in spans:
                label_data.append([start, end, label])

        # Tri optionnel des labels selon la position de début
        label_data.sort(key=lambda x: x[0])
        return json.dumps({
            "text": text,
            "label": label_data
        }, ensure_ascii=False)



class UimaCasStrategy(Strategy):
    """Stratégie pour créer des annotations au format UIMA CAS."""
    
    def __init__(self, type_name: str = 'custom.NamedEntity'):
        """
        Args:
            type_name: Nom du type d'annotation à créer dans le CAS
        """
        self.type_name = type_name
    
    def annotate(self, text: str, coordinate: Dict[Tuple[Tuple[int, int], ...], str]) -> Cas:
        """
        Crée un UIMA CAS avec les annotations fournies.
        
        Args:
            text: Le texte source
            coordinate: Dictionnaire {((start1, end1), (start2, end2), ...): '<LABEL>'}
                       Exemple: {((77, 95), (128, 151)): '<n>', ((14, 24),): '<DATE>'}
        
        Returns:
            Cas: Objet UIMA CAS avec les annotations
        """
        # Créer un CAS vide
        cas = Cas()
        
        # Ajouter le texte
        cas.sofa_string = text
        
        # Définir le type d'annotation
        typesystem = cas.typesystem
        
        # Créer un type personnalisé si nécessaire
        if not typesystem.contains_type(self.type_name):
            NamedEntityType = typesystem.create_type(
                self.type_name,
                supertypeName='uima.tcas.Annotation'
            )
            typesystem.create_feature(
                NamedEntityType,
                'label',
                'uima.cas.String'
            )
        
        # Récupérer le type
        NamedEntity = typesystem.get_type(self.type_name)
        
        # Parcourir le dictionnaire et ajouter les annotations
        for spans_tuple, label in coordinate.items():
            # Chaque spans_tuple contient un ou plusieurs (start, end)
            for start, end in spans_tuple:
                annotation = NamedEntity(begin=start, end=end, label=label)
                cas.add(annotation)
        return cas
    
    def save_to_json(self, cas: Cas, output_path: str):
        """Sauvegarde le CAS en JSON."""
        cas.to_json(output_path, pretty_print=True)
    
    def save_to_xmi(self, cas: Cas, output_path: str):
        """Sauvegarde le CAS en XMI (format XML UIMA standard)."""
        cas.to_xmi(output_path, pretty_print=True)



from incognito_anonymizer import PersonalInfo, anonymizer
import pytest
from datetime import datetime


def test_valid_personal_info():
    # Instance valide
    data = {
        "first_name": "John",
        "last_name": "Doe",
        "birth_name": "Smith",
        "birthdate": datetime(1990, 5, 15),
        "ipp": "123456789",
        "postal_code": "29200",
        "adress": "123 Main Street"
    }
    info = PersonalInfo(**data)

    assert info.first_name == "John"
    assert info.last_name == "Doe"
    assert info.birth_name == "Smith"
    assert info.birthdate == datetime(1990, 5, 15)
    assert info.ipp == "123456789"
    assert info.postal_code == "29200"
    assert info.adress == "123 Main Street"


def test_from_dict_valid_data():
    # Transformation d'un dictionnaire valide
    data = {
        "first_name": "Jane",
        "last_name": "Doe",
        "birth_name": "Johnson",
        "birthdate": datetime(1985, 8, 25),
        "ipp": "987654321",
        "postal_code": "75001",
        "adress": "456 Elm Street"
    }

    info = PersonalInfo(**data)
    assert info.first_name == "Jane"
    assert info.last_name == "Doe"
    assert info.birth_name == "Johnson"
    assert info.birthdate == datetime(1985, 8, 25)
    assert info.ipp == "987654321"
    assert info.postal_code == "75001"
    assert info.adress == "456 Elm Street"


def test_from_dict_missing_name():
    # Transformation d'un dictionnaire valide
    data = {
        "first_name": "Jane",
        "last_name": None,
        "birth_name": "Johnson",
        "birthdate": datetime(1985, 8, 25),
        "ipp": "987654321",
        "postal_code": "75001",
        "adress": "456 Elm Street"
    }
    ano = anonymizer.Anonymizer()
    info = ano.set_info_from_dict(**data)
    assert info.first_name == "Jane"
    assert info.last_name == ""
    assert info.birth_name == "Johnson"
    assert info.birthdate == datetime(1985, 8, 25)
    assert info.ipp == "987654321"
    assert info.postal_code == "75001"
    assert info.adress == "456 Elm Street"


def test_from_dict_full_date():
    # Transformation d'un dictionnaire valide
    data = {
        "first_name": "Jane",
        "last_name": None,
        "birth_name": "Johnson",
        "birthdate": datetime(1985, 8, 25, 0, 0, 0),
        "ipp": "987654321",
        "postal_code": "75001",
        "adress": "456 Elm Street"
    }
    ano = anonymizer.Anonymizer()
    info = ano.set_info_from_dict(**data)
    assert info.first_name == "Jane"
    assert info.last_name == ""
    assert info.birth_name == "Johnson"
    assert info.birthdate == datetime(1985, 8, 25)
    assert info.ipp == "987654321"
    assert info.postal_code == "75001"
    assert info.adress == "456 Elm Street"


def test_from_dict_missing_data():
    # Transformation avec données manquantes
    data = {
        "first_name": "Alice",
        "last_name": "Wonderland"
        # Les autres champs sont absents
    }
    info = PersonalInfo(**data)

    assert info.first_name == "Alice"
    assert info.last_name == "Wonderland"
    assert info.birth_name == ""
    assert info.birthdate == datetime(1000, 1, 1)
    assert info.ipp == ""
    assert info.postal_code == "0"
    assert info.adress == ""


def test_validation_error():
    # Données invalides
    with pytest.raises(ValueError):
        PersonalInfo(
            first_name="John",
            last_name="Doe",
            birthdate="invalid_date",  # Date invalide
            ipp="123456789"
        )



from incognito_anonymizer import anonymizer


def test_standoff():
    ano = anonymizer.Anonymizer()
    ano.set_annotator("standoff")
    ano.add_analyzer("regex")
    input_text = "Dr. Bob"
    output_text = ano.annotate(input_text)
    assert output_text == "T1\tNAME 4 7\tBob"


def test_doccano():
    ano = anonymizer.Anonymizer()
    ano.set_annotator("doccano")
    ano.add_analyzer("regex")
    input_text = "Dr. Bob"
    output_text = ano.annotate(input_text)
    assert output_text == '{"text": "Dr. Bob", "label": [[4, 7, "NAME"]]}'



import unicodedata
from datetime import datetime
from typing import Dict, Iterable, Optional, Tuple

import regex
import re
import warnings
from flashtext import KeywordProcessor
from pydantic import BaseModel


class PersonalInfo(BaseModel):
    first_name: str = ""
    last_name: str = ""
    birth_name: Optional[str] = ""
    birthdate: datetime = datetime(year=1000, month=1, day=1)
    ipp: str = ""
    iep: str = ""
    postal_code: Optional[str] = "0"
    adress: Optional[str] = ""


class AnalyzerStrategy:
    """Constructeur de la Class Strategy"""

    def analyze(text: str, info: PersonalInfo = None):
        raise NotImplementedError()


class PiiStrategy(AnalyzerStrategy):
    """Detect personal infos"""

    def __init__(self):
        pass

    def hide_by_keywords(
        self, text: str, keywords: Iterable[Tuple[str, str]]
    ) -> Dict[Tuple[int, int], str]:
        """
        Hide text using keywords and return positions with replacements.

        :param text: text to anonymize
        :param keywords: Iterable of tuples (word, replacement).


        :returns: List of tuples where each tuple contains:
                - A tuple with the start and end positions of the word.
                - The replacement string.
        """
        processor = KeywordProcessor(case_sensitive=False)
        for key, masks in keywords:
            key = "".join(
                (
                    c
                    for c in unicodedata.normalize("NFD", key)
                    if unicodedata.category(c) != "Mn"
                )
            )
            processor.add_keyword(key, masks)

        normalized_text = "".join(
            (
                c
                for c in unicodedata.normalize("NFD", text)
                if unicodedata.category(c) != "Mn"
            )
        )
        # Extract keywords with positions
        found_keywords = processor.extract_keywords(normalized_text, span_info=True)

        result = {}
        for replacement, start, end in found_keywords:
            # Wrap positions as a tuple of tuples
            key = ((start, end),)
            # if key in result:
            #     result[key] = replacement  # Handle multiple occurrences
            # else:
            result[key] = replacement
        return result

    def analyze(self, text: str, info: PersonalInfo = None) -> str:
        """
        Hide specific words based on keywords

        :param text: text to anonymize
        """
        keywords: tuple
        if not isinstance(info, PersonalInfo):
            print("info must be a Personnal info type. Returning empty dict instead.")
            return {}
        keywords = (
            (info.first_name, "<NAME>"),
            (info.last_name, "<NAME>"),
            (info.birth_name, "<NAME>"),
            (info.ipp, "<IPP>"),
            (info.iep, "<IEP>"),
            (info.postal_code, "<CODE_POSTAL>"),
            (info.birthdate.strftime("%m/%d/%Y"), "<DATE>"),
            (info.birthdate.strftime("%m %d %Y"), "<DATE>"),
            (info.birthdate.strftime("%m:%d:%Y"), "<DATE>"),
            (info.birthdate.strftime("%m-%d-%Y"), "<DATE>"),
            (info.birthdate.strftime("%Y-%m-%d"), "<DATE>"),
            (info.birthdate.strftime("%d/%m/%Y"), "<DATE>"),
            (info.adress, "<ADRESSE>"),
        )

        return self.hide_by_keywords(text, [(k, t) for k, t in keywords if k])


class RegexStrategy(AnalyzerStrategy):
    """Detect word based on regex"""

    def __init__(self):
        super().__init__
        Xxxxx = r"[A-ZÀ-Ÿ]\p{Ll}+"
        XXxX_ = r"[A-ZÀ-Ÿ][A-ZÀ-Ÿ\p{Ll}-]"
        XXxX_apostrophe = r"[A-ZÀ-Ÿ][A-ZÀ-Ÿ\p{Ll}-]*(?:[''][A-ZÀ-Ÿ][A-ZÀ-Ÿ\p{Ll}-]*)?"

        sep = r"(?:[ ]*|-)?"
        mois = r"(?i)(?:janvier|février|fevrier|mars|avril|mai|juin|juillet|août|aout|septembre|octobre|novembre|décembre|decembre|janv?[.]?|févr?[.]?|fevr?[.]?|avr[.]?|juil[.]?|sept?[.]?|oct[.]?|nov[.]?|déc[.]?|dec[.]?)"

        # Date complète littérale : "8 juillet 2020"
        self.date_litteral_full = (
            rf"\b(0?[1-9]|[12]\d|3[01])[\s]+{mois}[\s,]+((?:1[6-9]|[2-9]\d)\d{{2}})\b"
        )

        # Date partielle sans année : "20 mars"
        self.date_litteral_partial = rf"\b(0?[1-9]|[12]\d|3[01])[\s]+{mois}\b"

        # Mois seul : "juillet 2020" ou juste "juillet"
        self.mois_pattern = rf"\b{mois}(?:[\s]+((?:1[6-9]|[2-9]\d)\d{{2}}))?\b"
        self.title_regex = r"([Dd][Rr][.]?|[Dd]octeur|[mM]r?[.]?|[Ii]nterne[ ]*:?|INT|[Ee]xterne[ ]*:?|[Mm]onsieur|[Mm]adame|[Rr].f.rent[ ]*:?|[P][Rr][.]?|[Pp]rofesseure|[Pp]rofesseur|[Mm]me[.]?|[Ee]nfant|[Mm]lle|[Nn]ée?|[Cc]hef(fe)? de service|[Nn]om :)"

        self.email_pattern = (
            r"(?i)"
            r"(?:"
            r"[a-z0-9!#$%&'*+/=?^_`{|}~<>()\[\]\\:;,@\"\-]+"  # partie locale étendue
            r"(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~<>()\[\]\\:;,@\"\-]+)*"
            r"|"
            r"\"(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21\x23-\x5b\x5d-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])*\""
            r")"
            r"@"
            r"(?:"
            r"(?:[a-z0-9*<>()\[\]!#$%&'+=?^_`{|}~-](?:[a-z0-9*<>()\[\]!#$%&'+=?^_`{|}~-]*[a-z0-9*<>()\[\]!#$%&'+=?^_`{|}~-])?\.)*"
            r"[a-z0-9*<>()\[\]!#$%&'+=?^_`{|}~-](?:[a-z0-9*<>()\[\]!#$%&'+=?^_`{|}~-]*[a-z0-9*<>()\[\]!#$%&'+=?^_`{|}~-])?"
            r"|\[(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}"
            r"(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?|[a-z0-9-]*[a-z0-9]:"
            r"(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21-\x5a\x53-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])+)\]"
            r")"
        )
        # needs a comma  or \r to match. If it's in a middle of a phrase it won't match
        self.adresse_pattern = r"(?i)\d{1,4}\s*(?:bis|ter|quater)?\s+(?:rue|avenue|av\.|boulevard|bd\.?|impasse|allée|allee|chemin|route|place|square|résidence|residence|hameau|lieu[- ]dit|voie|passage|villa|domaine|lotissement|parc|traverse|ruelle|sentier|cours|quai|esplanade)\s+[a-z0-9éèàùâêîôûïëüçæœ'\-\.]+(?:\s+[a-z0-9éèàùâêîôûïëüçæœ'\-\.]+){0,10},?\s*\d{5},?\s*[a-zéèàùâêîôûïëüçæœ'\-\.]+(?:\s+[a-zéèàùâêîôûïëüçæœ'\-\.]+){0,5}(?=\s*[,\{\n]|$)"

        # INFO: Non restrictive regexp for matching 3 word after a street description.
        self.fast_adresse_pattern = r"(?i)(?:\d+\s+)?(rue|avenue|av|boulevard|bd|bld|allée|allee|impasse|chemin|route|place|square|villa|passage|domaine|hameau|lotissement|résidence|residence|quartier|sentier|traverse|cours|quai|esplanade|promenade|rond[- ]point)\b(?:\s+\S+){1,3}"

        self.zip_city_name = (
            r"\b(\d{5})\s+([A-ZÀÂÉÈÊËÎÏÔÙÛÜÇ][A-ZÀÂÉÈÊËÎÏÔÙÛÜÇ\s\-]+)\b"
        )
        self.PATTERNS = {
            # rf"(?<={self.title_regex})([\s-][A-Z]+)+([\s-][A-Z][a-z]+)+(?![a-z])": "<NAME>",
            rf"(?<={self.title_regex}[ \n]+)(?P<LN0>[A-ZÀ-Ÿ][A-ZÀ-Ÿ](?:{sep}(?:ep[.]|de|[A-ZÀ-Ÿ]+))*)[ ]+(?P<FN0>{Xxxxx}(?:{sep}{Xxxxx})*)": "<NAME>",
            rf"(?<={self.title_regex}[ \n]+)(?P<FN1>{Xxxxx}(?:{sep}{Xxxxx})*)[ ]+(?P<LN1>[A-ZÀ-Ÿ][A-ZÀ-Ÿ]+(?:{sep}(?:ep[.]|de|[A-ZÀ-Ÿ]+))*)": "<NAME>",
            rf"(?<={self.title_regex}[ \n]+)(?P<LN3>{Xxxxx}(?:(?:-|[ ]de[ ]|[ ]ep[.][ ]){Xxxxx})*)[ ]+(?P<FN2>{Xxxxx}(?:-{Xxxxx})*)": "<NAME>",
            rf"(?<={self.title_regex}[ \n]+)(?P<LN2>{XXxX_}+(?:{sep}{XXxX_}+)*)": "<NAME>",
            rf"(?<={self.title_regex}[ \n]+)(?P<FN0>[A-ZÀ-Ÿ][.])\s+(?P<LN0>{XXxX_}+(?:{sep}{XXxX_}+)*)": "<NAME>",
            rf"(?<={self.title_regex}[ \n]+)(?P<FN0>[A-ZÀ-Ÿ][.](?:[A-ZÀ-Ÿ][.])*)\s+(?P<LN0>{XXxX_apostrophe}+(?:{sep}{XXxX_apostrophe}+)*)": "<NAME>",
            rf"(?<={self.title_regex}[ \n]+)(?P<FN0>[A-ZÀ-Ÿ][.](?:[A-ZÀ-Ÿ][.])*)\s+(?:de |d'|du |des )?(?P<LN0>{XXxX_apostrophe}+(?:{sep}{XXxX_apostrophe}+)*)": "<NAME>",
            # r"[12]\s*[0-9]{2}\s*(0[1-9]|1[0-2])\s*(2[AB]|[0-9]{2})\s*[0-9]{3}\s*[0-9]{3}\s*(?:\(?([0-9]{2})\)?)?": "<NIR>",
            # r"(?:(?:\+|00)33[\s.-]*|0)[\s.-]*[1-9](?:[\s.-]*\d{2}){4}": "<PHONE>",
            self.date_litteral_full: "<DATE>",  # 8 juillet 2020  ← plus spécifique en premier
            self.date_litteral_partial: "<DATE>",  # 20 mars
            self.mois_pattern: "<DATE>",
            r"\b(0?[1-9]|[12]\d|3[01])(\/|-|\.)(0?[1-9]|1[0-2])\2((?:(?:1[6-9]|[2-9]\d)\d{2}|\d{2}))\b": "<DATE>",
            self.email_pattern: "<EMAIL>",
            self.adresse_pattern: "<ADRESSE>",
            self.zip_city_name: "<ADRESSE>",
            self.mois_pattern: "<DATE>",
            r"(?:(?:\+|00)33[\s.-]*|0)[\s.-]*[1-9](?:[\s.-]*\d{2}){4}|\(?\d[\d\s]{6,}\d": "<NUMBER>",
            self.fast_adresse_pattern: "<ADRESSE>",
            # r"[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ]{4,}\s+[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{4,}\s": "<NAME>",
        }

    def multi_subs_by_regex(self, text: str) -> Dict[Tuple[Tuple[int, int]], str]:
        """
        Find word position based on regex

        :param text: text to anonymise
        :returns: List of tuples where each tuple contains:
                - A tuple with the start and end positions of the word.
                - The replacement string.
        """

        self.position = {}

        for pattern, repl in self.PATTERNS.items():
            matches_iter = list(regex.finditer(pattern, text, overlapped=True))
            if not matches_iter:
                continue

            spans = [match.span() for match in matches_iter]

            # Dédoublonnage : pour les spans overlappants, garder uniquement le plus long
            filtered_spans = self._remove_overlapping_spans(spans)
            existing_keys = list(self.position.keys())
            overlapping_keys = []
            for key in existing_keys:
                if any(span in key for span in filtered_spans) or any(
                    k in filtered_spans for k in key
                ):
                    overlapping_keys.append(key)

            if overlapping_keys:
                combined_key = tuple(
                    sorted(
                        set(span for key in overlapping_keys for span in key).union(
                            filtered_spans
                        )
                    )
                )
                for key in overlapping_keys:
                    del self.position[key]
                self.position[combined_key] = repl
            else:
                self.position[tuple(filtered_spans)] = repl

        result = {}
        for k, v in self.position.items():
            if v != "<EMAIL>":
                result[k] = v
                continue

            email_tuples = list(k)
            ends = {}

            for start, end in email_tuples:
                length = end - start
                if end not in ends or length > (ends[end][1] - ends[end][0]):
                    ends[end] = (start, end)

            result[tuple(ends.values())] = "<EMAIL>"

        self.position = self._resolve_position_conflicts(result)
        return self.position

    def analyze(self, text: str, info: PersonalInfo = None):
        """
        Hide text using regular expression
        :param text: text to anonymize
        """
        return self.multi_subs_by_regex(text)

    def _remove_overlapping_spans(self, spans: list) -> list:
        """
        Pour un ensemble de spans potentiellement overlappants,
        ne garder que les spans non-overlappants les plus longs.
        """
        if not spans:
            return spans

        # Trier par longueur décroissante (garder les plus longs en priorité)
        sorted_spans = sorted(spans, key=lambda s: s[1] - s[0], reverse=True)

        kept = []
        for span in sorted_spans:
            start, end = span
            # Vérifier si ce span overlap avec un span déjà gardé
            overlaps = any(
                not (end <= kept_start or start >= kept_end)
                for kept_start, kept_end in kept
            )
            if not overlaps:
                kept.append(span)

        # Retrier par position
        return sorted(kept, key=lambda s: s[0])

    def _spans_overlap(self, span1: Tuple[int, int], span2: Tuple[int, int]) -> bool:
        """Vérifie si deux spans se chevauchent"""
        return not (span1[1] <= span2[0] or span2[1] <= span1[0])

    def _resolve_position_conflicts(
        self, positions: Dict[Tuple[Tuple[int, int]], str]
    ) -> Dict[Tuple[Tuple[int, int]], str]:
        """
        Pour des clés de position qui se chevauchent avec la même valeur,
        ne garder que la clé avec le span le plus large.

        :param positions: dict avec des tuples de spans comme clés et des remplacements comme valeurs
        :returns: dict filtré sans conflits de positions
        """
        result = dict(positions)
        keys = list(result.keys())
        to_delete = set()

        for i, key1 in enumerate(keys):
            if key1 in to_delete:
                continue
            for key2 in keys[i + 1 :]:
                if key2 in to_delete:
                    continue
                if result[key1] != result[key2]:
                    continue

                has_overlap = any(
                    self._spans_overlap(s1, s2) for s1 in key1 for s2 in key2
                )

                if has_overlap:
                    len1 = max(end - start for start, end in key1)
                    len2 = max(end - start for start, end in key2)
                    to_delete.add(key1 if len1 < len2 else key2)

        for key in to_delete:
            del result[key]

        return result


class LossyStrategy(RegexStrategy):
    """
    Find word position based on regex

    :param text: text to anonymise
    :returns: List of tuples where each tuple contains:
            - A tuple with the start and end positions of the word.
            - The replacement string.
    .. warning::
    This strategy is intentionally lossy: it trades recall precision for
    maximum anonymization coverage. Information loss is expected and assumed.
    """

    def __init__(self):
        super().__init__
        self.title_regex = r"([Dd][Rr][.]?|[Dd]octeur|[mM]r?[.]?|[Ii]nterne[ ]*:?|INT|[Ee]xterne[ ]*:?|[Mm]onsieur|[Mm]adame|[Rr].f.rent[ ]*:?|[P][Rr][.]?|[Pp]rofesseure|[Pp]rofesseur|[Mm]me[.]?|[Ee]nfant|[Mm]lle|[Nn]ée?|[Cc]hef(fe)? de service|[Nn]om :)"
        self.LOSSY_PATTERNS = {
            # DUPONT Martin ou DUPONT de TOTO Martin ou DUPONT-TOTO Martin
            r"([A-Z][A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ]*){2,}([ \t]+([A-Z][A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ]*|de|du|des|von|van|le|la)){0,3}[ \t]+[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{2,}(-[A-Z][a-z-éèçùàâêîôûëïü]{2,})*": "<NAME>",
            # Martin DUPONT ou Martin DUPONT de TOTO ou Martin DUPONT-TOTO
            r"[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{2,}(-[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{2,})*[ \t]+([A-Z][A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ]*){2,}([ \t]+([A-Z][A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ]*|de|du|des|von|van|le|la)){0,3}": "<NAME>",
            # J. Pierre ou J.P. Marie
            r"([A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ]\.){1,3}[ \t]*[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{2,}(-[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{2,})*": "<NAME>",
            # DUPONT Jean-Philippe ou DUPONT Jean Philippe (prénom composé avec ou sans trait d'union)
            r"([A-Z][A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ]*){2,}([ \t]+([A-Z][A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ]*|de|du|des|von|van|le|la)){0,3}[ \t]+[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{2,}(-[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{2,})*([ \t]+[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{2,}(-[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{2,})*)+": "<NAME>",
            # L Philippe ou L. Philippe (initiale suivie d'un prénom)
            r"\b[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ]\.?[ \t]+[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{2,}(-[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{2,})*": "<NAME>",
            # Philippe LOC'H (prénom suivi d'un nom avec apostrophe)
            r"[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{2,}(-[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{2,})*[ \t]+([A-Z][A-Z'-ÉÈÀÂÊÎÔÛËÏÜÙÇ]*){2,}": "<NAME>",
            # DUPONT Martin ou DUPONT de TOTO Martin
            rf"(?:{self.title_regex}[ \t\n]+)?([A-Z][A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ]*){"{2,}"}([ \t]+([A-Z][A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ]*|de|du|des|von|van|le|la)){{0,3}}[ \t]+[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{{2,}}(-[A-Z][a-z-éèçùàâêîôûëïü]{{2,}})*": "<NAME>",
            # Martin DUPONT
            rf"(?:{self.title_regex}[ \t\n]+)?[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{{2,}}(-[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{{2,}})*[ \t]+([A-Z][A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ]*){{2,}}([ \t]+([A-Z][A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ]*|de|du|des|von|van|le|la)){{0,3}}": "<NAME>",
            # J. Pierre ou J.P. Marie
            rf"(?:{self.title_regex}[ \t\n]+)?([A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ]\.){{1,3}}[ \t]*[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{{2,}}(-[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{{2,}})*": "<NAME>",
            # DUPONT Jean-Philippe
            rf"(?:{self.title_regex}[ \t\n]+)?([A-Z][A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ]*){{2,}}([ \t]+([A-Z][A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ]*|de|du|des|von|van|le|la)){{0,3}}[ \t]+[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{{2,}}(-[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{{2,}})*([ \t]+[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{{2,}}(-[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{{2,}})*)+": "<NAME>",
            # L. Philippe
            rf"(?:{self.title_regex}[ \t\n]+)?\b[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ]\.?[ \t]+[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{{2,}}(-[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{{2,}})*": "<NAME>",
            # Philippe LOC'H
            rf"(?:{self.title_regex}[ \t\n]+)?[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{{2,}}(-[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ][a-z-éèçùàâêîôûëïü]{{2,}})*[ \t]+([A-Z][A-Z'-ÉÈÀÂÊÎÔÛËÏÜÙÇ]*){{2,}}": "<NAME>",
            # B. ALBERT (initiale + point + nom en majuscules)
            rf"(?:{self.title_regex}[ \t\n]+)?[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ]\.[ \t]+[A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ]{{2,}}([A-Z-ÉÈÀÂÊÎÔÛËÏÜÙÇ]*)": "<NAME>",

        }
    def multi_subs_by_regex(self, text: str) -> Dict[Tuple[Tuple[int, int]], str]:
        """
        Analyze text using an aggressive uppercase-based matching strategy.

        :param text: Text to anonymize.
        :returns: Dictionary mapping span tuples to replacement strings.

        .. warning::
            This strategy can suppress legitimate content. Any token matching
            the uppercase pattern will be replaced, regardless of whether it
            is actually a personal identifier.
        """

        self.position = {}
        text = text.replace('\x7f', '')
        for pattern, repl in self.LOSSY_PATTERNS.items():
            matches_iter = list(regex.finditer(pattern, text, overlapped=True))
            if not matches_iter:
                continue

            spans = [match.span() for match in matches_iter]
            filtered_spans = self._remove_overlapping_spans(spans)
            existing_keys = list(self.position.keys())

            overlapping_keys = [
                key
                for key in existing_keys
                if any(span in key for span in filtered_spans)
                or any(k in filtered_spans for k in key)
            ]

            if overlapping_keys:
                combined_key = tuple(
                    sorted(
                        set(span for key in overlapping_keys for span in key).union(
                            filtered_spans
                        )
                    )
                )
                for key in overlapping_keys:
                    del self.position[key]
                self.position[combined_key] = repl
            else:
                self.position[tuple(filtered_spans)] = repl

        self.position = self._resolve_position_conflicts(self.position)
        return self.position

    def analyze(self, text: str, info: PersonalInfo = None):
        """
        Hide text using regular expression
        :param text: text to anonymize
        """
        warnings.warn(
            "LossyStrategy.analyze() uses aggressive pattern matching that may cause "
            "unintended information loss. Tokens matching the uppercase pattern will be "
            "replaced unconditionally, including potential false positives such as "
            "acronyms, place names, or medical terminology. "
            "Use a more precise strategy if data integrity is critical.",
            UserWarning,
            stacklevel=2,
        )
        return self.multi_subs_by_regex(text)
